using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.SqlServer.Management.UI.VSIntegration;
using Microsoft.SqlServer.Management.UI.VSIntegration.Editors;
using Microsoft.SqlServer.Management.Smo.RegSvrEnum;
using Microsoft.SqlServer.Management.Common;
using Microsoft.SqlServer.Management.Smo;
using System.Reflection;
using System.Windows;
using EnvDTE80;
using EnvDTE;
using Microsoft.SqlServer.Management.UI.VSIntegration.ObjectExplorer;
using System.Threading;
using System.Text.RegularExpressions;


namespace DatabaseObjectSearcher
{
    internal class ObjectExplorerManager
    {
        

        internal void SelectSMOObjectInObjectExplorer(NamedSmoObject objectToSelect, SqlConnectionInfo connection)
        {
            IExplorerHierarchy hierarchy = GetHierarchyForConnection(connection);
            if (hierarchy == null)
            {
                return; // there is nothing we can really do if we don't have one of these
            }
            HierarchyTreeNode databasesNode = GetUserDatabasesNode(hierarchy.Root);
            SelectSMOObject(databasesNode, objectToSelect);
        }

        private HierarchyTreeNode GetUserDatabasesNode(HierarchyTreeNode rootNode)
        {
            if (rootNode != null)
            {
                // root should always be expandable
                if (rootNode.Expandable)
                {
                    EnumerateChildrenSynchronously(rootNode);
                    rootNode.Expand();

                    // TODO this is horrible code - it assumes the first node will ALWAYS be the "Databases" node in the object explorer, which may not always be the case
                    // however I couldn't think of a clean way to always find the right node
                    return (HierarchyTreeNode)rootNode.Nodes[0];
                }
            }
            return null;
        }

        private string BuildMatchingPathExpressionForDepth(NamedSmoObject objectToSelect, string parentNodePath, int level, out bool atFinalLevel)
        {
            atFinalLevel = false;
            string expression = string.Empty;

            // Databases node is at level 1
            switch (level)
            {
                case 2: // database level
                    Regex re = new Regex(".*Database\\[@Name='(.*?)']?");
                    Match m = re.Match(objectToSelect.Urn);
                    expression = parentNodePath + "\\" + m.Groups[1].Captures[0];
                    atFinalLevel = (objectToSelect is Database);
                    break;

                case 3:
                    if (objectToSelect is StoredProcedure)
                    {
                        expression = parentNodePath + "\\Programmability";
                    }
                    else if (objectToSelect is Table)
                    {
                        expression = parentNodePath + "\\Tables";
                    }
                    else if (objectToSelect is Microsoft.SqlServer.Management.Smo.View)
                    {
                        expression = parentNodePath + "\\Views";
                    }
                    break;

                case 4:
                    if (objectToSelect is StoredProcedure)
                    {
                        expression = parentNodePath + "\\Stored Procedures";
                    }
                    else
                    {
                        expression = parentNodePath + "\\" + GetSchemaQualifiedNameForSmoObject(objectToSelect);
                        atFinalLevel = true;
                    }
                    break;

                case 5:
                    expression = parentNodePath + "\\" + GetSchemaQualifiedNameForSmoObject(objectToSelect);
                    atFinalLevel = true;
                    break;
            }

            return expression;
        }

        private string GetSchemaQualifiedNameForSmoObject(NamedSmoObject namedObject)
        {
            Regex re = new Regex(".*\\[@Name='(.*?)' and @Schema='(.*?)']");
            Match m = re.Match(namedObject.Urn);
            string schemaQualifiedName = m.Groups[2].Captures[0] + "." + m.Groups[1].Captures[0];
            return schemaQualifiedName; // Named SMO object has a FullQualifiedName property but it is internal
        }

        private void SelectSMOObject(HierarchyTreeNode node, NamedSmoObject objectToSelect)
        {
            if (node != null)
            {
                if (node.Expandable)
                {
                    EnumerateChildrenSynchronously(node);
                    node.Expand();

                    bool atFinalLevel;
                    string pattern = BuildMatchingPathExpressionForDepth(objectToSelect, node.FullPath, node.Level + 1, out atFinalLevel);

                    foreach (HierarchyTreeNode child in node.Nodes)
                    {
                        if (string.Compare(child.FullPath, pattern, true) == 0)
                        {
                            if (atFinalLevel)
                            {
                                SelectNode(child);
                            }
                            else
                            {
                                SelectSMOObject(child, objectToSelect);
                            }
                        }
                    }
                }
            }
        }


        private void SelectNode(HierarchyTreeNode node)
        {
            IServiceProvider provider = node as IServiceProvider;
            if (provider != null)
            {
                INodeInformation containedItem = provider.GetService(typeof(INodeInformation)) as INodeInformation;
                if (containedItem != null)
                {
                    IObjectExplorerService objExplorer = ServiceCache.GetObjectExplorer();
                    objExplorer.SynchronizeTree(containedItem);
                }
            }
        }

        // ugly reflection hack
        private IExplorerHierarchy GetHierarchyForConnection(SqlConnectionInfo connection)
        {
            IObjectExplorerService objExplorer = ServiceCache.GetObjectExplorer();
            Type t = objExplorer.GetType();
            MethodInfo getHierarchyMethod = t.GetMethod("GetHierarchy", BindingFlags.Instance | BindingFlags.NonPublic);
            if (getHierarchyMethod != null)
            {
                IExplorerHierarchy hierarchy = getHierarchyMethod.Invoke(objExplorer, new object[] { connection }) as IExplorerHierarchy;
                return hierarchy;
            }

            return null;
        }

        // another exciting opportunity to use reflection
        private void EnumerateChildrenSynchronously(HierarchyTreeNode node)
        {
            Type t = node.GetType();
            MethodInfo method = t.GetMethod("EnumerateChildren", new Type[] { typeof(Boolean) });
            if (method != null)
            {
                method.Invoke(node, new object[] { false });
            }
            else
            {
                // fail
                node.EnumerateChildren();
            }
        }
    }
}
